"""ASpecD framework

Tools for reproducible and traceable analysis of spectral data.

Available modules
-----------------
:mod:`aspecd.dataset`
    Unit containing data and metadata
:mod:`aspecd.metadata`
    Organise metadata in a dataset
:mod:`aspecd.io`
    Import and export data into and from datasets
:mod:`aspecd.infofile`
    Import and parse info files
:mod:`aspecd.processing`
    Process data in datasets
:mod:`aspecd.analysis`
    Analyse data in datasets
:mod:`aspecd.annotation`
    Annotate datasets and plots
:mod:`aspecd.plotting`
    Plot data in datasets
:mod:`aspecd.table`
    Tabulate data in datasets
:mod:`aspecd.history`
    Collect information on what has been done to a dataset
:mod:`aspecd.report`
    Reports textually presenting the information accumulated
:mod:`aspecd.tasks`
    Recipe-driven data analysis
:mod:`aspecd.model`
    Numerical models

Utilities
---------
:mod:`aspecd.system`
    Information about the system used
:mod:`aspecd.utils`
    Basic functionality used by several modules.
:mod:`aspecd.exceptions`
    Exceptions for the ASpecD package

"""
